@@ -1,6 +1,7 @@
 """Core classes used to power pipelines"""
 
-from collections import OrderedDict
+import asyncio
+from collections import defaultdict, OrderedDict
 from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
 import copy
 
@@ -10,7 +11,6 @@ except ImportError:
     import profile
 from inspect import signature, Parameter
 import os
-from pprint import pformat
 
 from consecution import (
     Pipeline,
@@ -21,13 +21,14 @@ from numpydoc.docscrape import FunctionDoc
 from tlbx import (
     st,
     repr,
+    pf,
     Script,
     Arg,
     Parent,
     MappingMixin,
     set_missing_key,
     format_msg,
-    shell as shell_cmd
+    shell as shell_cmd,
 )
 
 from glide.utils import (
@@ -43,10 +44,18 @@ from glide.utils import (
     load_json_config,
     load_ini_config,
     load_yaml_config,
+    get_or_create_event_loop,
+    cancel_asyncio_tasks,
 )
 
-SCRIPT_DATA_ARG = "data"
+SCRIPT_DATA_ARG = "glide_data"
+RESERVED_ARG_NAMES = [SCRIPT_DATA_ARG]
+# These are reserved because they are arg names used by consume(). Since node
+# context can be passed to consume() as node_name->arg_dict pairs in consume's
+# kwargs, we prevent nodes from using these names to avoid conflicts.
 RESERVED_NODE_NAMES = set(["cleanup", "split_count", "synchronous", "timeout"])
+# A backup timeout when trying to cancel unfinished async tasks
+ASYNCIO_CANCEL_TIMEOUT = 0.1
 
 
 class PushTypes:
@@ -174,6 +183,10 @@ class Node(ConsecutionNode):
                 # directly in process()
                 continue
 
+            assert param.name not in RESERVED_ARG_NAMES, (
+                "Reserved arg name '%s' used in run()" % param.name
+            )
+
             if param.kind == param.POSITIONAL_ONLY:
                 positionals[param.name] = None
             elif (
@@ -423,7 +436,7 @@ class PoolSubmit(Node):
         timeout : int or float, optional
             Time to wait for jobs to complete before raising an error. Ignored
             unless using a push_type that waits for results.
-        push_type : type, optional
+        push_type : str, optional
             If "async", push the Futures immediately.
             If "input", push the input data immediately after task submission.
             If "result", collect the result synchronously and push it.
@@ -502,6 +515,71 @@ class ThreadPoolSubmit(ProcessPoolSubmit):
         return ThreadPoolExecutor(**executor_kwargs)
 
 
+class AsyncIOSubmit(Node):
+    """A node that splits input data over an async function"""
+
+    def get_results(self, futures, timeout=None):
+        loop = asyncio.get_event_loop()
+        done, pending = loop.run_until_complete(asyncio.wait(futures, timeout=timeout))
+        if timeout and pending:
+            cancel_asyncio_tasks(pending, loop, cancel_timeout=ASYNCIO_CANCEL_TIMEOUT)
+            raise asyncio.TimeoutError(
+                "%d/%d tasks unfinished" % (len(pending), len(futures))
+            )
+        results = [task.result() for task in done]
+        return flatten(results)
+
+    def run(
+        self, data, func, split_count=None, timeout=None, push_type=PushTypes.Async
+    ):
+        """Use a asyncio to apply func to data
+
+        Parameters
+        ----------
+        data
+            An iterable to process
+        func : callable
+            A async callable that will be passed data to operate on using asyncio.
+        split_count : int, optional
+            How many slices to split the data into for concurrent processing. Default
+            is to set split_count = len(data).
+        timeout : int or float, optional
+            Time to wait for jobs to complete before raising an error. Ignored
+            unless using a push_type that waits for results.
+        push_type : str, optional
+            If "async", push the Futures immediately.
+            If "input", push the input data immediately after task submission.
+            If "result", collect the result synchronously and push it.
+
+        """
+        split_count = split_count or len(data)
+        splits = divide_data(data, split_count)
+        info(
+            "%s: data len: %s, splits: %s"
+            % (self.__class__.__name__, size(data, "n/a"), split_count)
+        )
+
+        loop, close = get_or_create_event_loop()
+
+        try:
+            futures = [loop.create_task(func(split)) for split in splits]
+
+            if push_type == PushTypes.Async:
+                for future in futures:
+                    self.push(future)
+            elif push_type == PushTypes.Input:
+                self.push(data)
+            elif push_type == PushTypes.Result:
+                self.push(self.get_results(futures, timeout=timeout))
+            else:
+                assert False, "Invalid push_type: %s" % push_type
+        finally:
+            if close and push_type == PushTypes.Result:
+                # We can only be sure its safe to close the event loop if it
+                # was created and all processing took place in here.
+                loop.close()
+
+
 class AssertFunc(Node):
     def run(self, data, func):
         """Assert that a function returns a truthy value
@@ -655,6 +733,14 @@ def clean_up_nodes(cleanup, contexts):
         if not ctx_value:
             dbg("Skipping clean up for %s->%s, value is blank" % (node_name, arg_name))
             continue
+
+        if isinstance(ctx_value, RuntimeContext):
+            dbg(
+                "Skipping clean up for %s->%s, value is RuntimeContext object"
+                % (node_name, arg_name)
+            )
+            continue
+
         dbg("Executing clean up for %s->%s" % (node_name, arg_name))
         try:
             func(ctx_value)
@@ -690,7 +776,7 @@ def consume(pipeline, data, cleanup=None, **node_contexts):
     update_node_contexts(pipeline, node_contexts)
     try:
         contexts = get_node_contexts(pipeline)
-        dbg("size=%s\n%s" % (size(data, "n/a"), pformat(contexts)), indent="label")
+        dbg("size=%s\n%s" % (size(data, "n/a"), pf(contexts)), indent="label")
         try:
             if data is None:
                 return consume_none(pipeline)
@@ -978,6 +1064,7 @@ class GliderScript(Script):
     ):
         """Generate the script args for the given Glider and return a decorator"""
         self.glider = glider
+        self.custom_args = script_args or []
         self.blacklist = set(blacklist or [])
 
         self.parents = parents or []
@@ -999,13 +1086,20 @@ class GliderScript(Script):
                 self.cleanup, dict
             ), "cleanup must be a dict of argname->func mappings"
 
-        script_args = self._get_script_args(script_args)
-        return super().__init__(*script_args)
+        self._check_arg_conflicts()
+        all_script_args = self._get_script_args()
+        return super().__init__(*all_script_args)
 
     def __call__(self, func, *args, **kwargs):
         func = self._node_arg_converter(func, *args, **kwargs)
         return super().__call__(func, *args, **kwargs)
 
+    def _check_arg_conflicts(self):
+        for dest in self._get_custom_arg_dests():
+            assert dest not in self.inject, (
+                "Arg dest '%s' conflicts with injected arg" % dest
+            )
+
     def get_injected_kwargs(self):
         """Override Script method to return populated kwargs from inject arg"""
         if not self.inject:
@@ -1052,6 +1146,48 @@ class GliderScript(Script):
             return True
         return False
 
+    def _get_custom_arg_dests(self):
+        return [a.dest for a in self.custom_args]
+
+    def _get_node_name_arg_map(self):
+        node_map = defaultdict(set)
+        node_lookup = self.glider.get_node_lookup()
+        for node in node_lookup.values():
+            for arg_name, _ in node.run_args.items():
+                node_map[node.name].add(arg_name)
+            for kwarg_name, kwarg_default in node.run_kwargs.items():
+                node_map[node.name].add(kwarg_name)
+        return node_map
+
+    def _get_arg_name_node_map(self):
+        arg_map = defaultdict(set)
+        node_lookup = self.glider.get_node_lookup()
+        for node in node_lookup.values():
+            for arg_name, _ in node.run_args.items():
+                arg_map[arg_name].add(node.name)
+            for kwarg_name, kwarg_default in node.run_kwargs.items():
+                arg_map[kwarg_name].add(node.name)
+        return arg_map
+
+    def _get_arg_node_name(self, arg):
+        names = []
+        node_lookup = self.glider.get_node_lookup()
+        for node in node_lookup.values():
+            if not arg.startswith(node.name + "_"):
+                continue
+
+            arg_base_name = arg[len(node.name) + 1 :]
+            if arg_base_name in node.run_args or arg_base_name in node.run_kwargs:
+                names.append(node.name)
+
+        assert len(names) <= 1, "More than one node found for arg name %s: %s" % (
+            arg,
+            names,
+        )
+        if not names:
+            return None
+        return names[0]
+
     def _get_script_arg_name(self, node_name, arg_name):
         return "%s_%s" % (node_name, arg_name)
 
@@ -1077,6 +1213,8 @@ class GliderScript(Script):
             arg_type = type(default)
 
         if arg_type == bool:
+            assert not required, "Required bool args don't make sense"
+
             base_arg_name = arg_name
             if default:
                 action = "store_false"
@@ -1097,6 +1235,10 @@ class GliderScript(Script):
             )
         else:
             arg_name = self._get_script_arg_name(node.name, arg_name)
+            # TODO: argparse puts required args with "--" in the "optional"
+            # section. There are workarounds, but it's unclear how to use them
+            # with tlbx.Arg which is based on the climax library.
+            # https://stackoverflow.com/q/24180527/10682164
             script_arg = Arg(
                 "--" + arg_name,
                 required=required,
@@ -1106,16 +1248,26 @@ class GliderScript(Script):
             )
         return script_arg
 
-    def _get_script_args(self, custom_script_args=None):
+    def _get_script_args(self):
         """Generate all tlbx Args for this Glider"""
         node_lookup = self.glider.get_node_lookup()
-        custom_script_args = custom_script_args or []
-        script_args = OrderedDict()
+        script_args = OrderedDict()  # Map of arg names to Args
+        arg_dests = {}  # Map of arg dests back to names
+        node_arg_names = defaultdict(set)
 
         requires_data = not isinstance(self.glider.top_node, NoInputNode)
         if requires_data and not self.blacklisted("", SCRIPT_DATA_ARG):
             script_args[SCRIPT_DATA_ARG] = Arg(SCRIPT_DATA_ARG, nargs="+")
 
+        def add_script_arg(node, arg_name, **kwargs):
+            script_arg = self._get_script_arg(node, arg_name, **kwargs)
+            if not script_arg:
+                return
+
+            script_args[script_arg.name] = script_arg
+            arg_dests[script_arg.dest] = script_arg.name
+            node_arg_names[arg_name].add(script_arg.name)
+
         for node in node_lookup.values():
             try:
                 # Only works if run() has docs in numpydoc format
@@ -1126,32 +1278,49 @@ class GliderScript(Script):
                 node_help = {}
 
             for arg_name, _ in node.run_args.items():
-                arg_help = node_help.get(arg_name, None)
-                script_arg = self._get_script_arg(
-                    node, arg_name, required=True, arg_help=arg_help
+                add_script_arg(
+                    node,
+                    arg_name,
+                    required=True,
+                    arg_help=node_help.get(arg_name, None),
                 )
-                if not script_arg:
-                    continue
-                script_args[script_arg.name] = script_arg
 
             for kwarg_name, kwarg_default in node.run_kwargs.items():
-                arg_help = node_help.get(kwarg_name, None)
-                script_arg = self._get_script_arg(
+                add_script_arg(
                     node,
                     kwarg_name,
                     required=False,
                     default=kwarg_default,
-                    arg_help=arg_help,
+                    arg_help=node_help.get(kwarg_name, None),
                 )
-                if not script_arg:
-                    continue
-                script_args[script_arg.name] = script_arg
 
-        for custom_arg in custom_script_args:
+        def assert_arg_present(custom_arg, arg_name):
+            assert arg_name in script_args, (
+                "Custom arg %s with dest=%s maps to node arg=%s "
+                "which is not in the script arg list. Check for "
+                "conflicting args that cover the same node arg."
+                % (custom_arg.name, custom_arg.dest, arg_name)
+            )
+
+        for custom_arg in self.custom_args:
             assert not self.blacklisted("", custom_arg.name), (
                 "Blacklisted arg '%s' passed as a custom arg" % custom_arg.name
             )
+
+            if custom_arg.dest in node_arg_names:
+                # Find and delete all node-based args this will cover
+                for arg_name in node_arg_names[custom_arg.dest]:
+                    assert_arg_present(custom_arg, arg_name)
+                    del script_args[arg_name]
+
+            if custom_arg.dest in arg_dests:
+                # Remove the original arg that this custom arg will satisfy
+                arg_name = arg_dests[custom_arg.dest]
+                assert_arg_present(custom_arg, arg_name)
+                del script_args[arg_name]
+
             script_args[custom_arg.name] = custom_arg
+            arg_dests[custom_arg.dest] = custom_arg.name
 
         return script_args.values()
 
@@ -1191,20 +1360,16 @@ class GliderScript(Script):
         add_to_final = set()
 
         for key, value in kwargs.items():
-            key_parts = key.split("_")
-            node_name = key_parts[0]
+            assert (
+                key not in nodes
+            ), "Invalid keyword arg '%s', can not be a node name" % (key)
 
+            node_name = self._get_arg_node_name(key)
             if node_name not in nodes:
                 add_to_final.add(key)
                 continue
 
-            if key in self.inject:
-                add_to_final.add(key)
-
-            assert (
-                len(key_parts) > 1
-            ), "Invalid keyword arg %s, can not be a node name" % (key)
-            arg_name = "_".join(key_parts[1:])
+            arg_name = key[len(node_name) + 1 :]
             node_contexts.setdefault(node_name, {})[arg_name] = value
 
         injected_node_contexts = self._get_injected_node_contexts(kwargs)
@@ -1214,6 +1379,18 @@ class GliderScript(Script):
             else:
                 node_contexts[node_name] = injected_args
 
+        arg_node_map = self._get_arg_name_node_map()
+        for custom_arg_dest in self._get_custom_arg_dests():
+            if custom_arg_dest not in kwargs:
+                continue
+
+            for node_name in arg_node_map.get(custom_arg_dest, []):
+                custom_arg_dict = {custom_arg_dest: kwargs[custom_arg_dest]}
+                if node_name in node_contexts:
+                    node_contexts[node_name].update(custom_arg_dict)
+                else:
+                    node_contexts[node_name] = custom_arg_dict
+
         final_kwargs = dict(node_contexts=node_contexts)
         for key in add_to_final:
             final_kwargs[key] = kwargs[key]
