@@ -1,6 +1,7 @@
 """Common utilities"""
 
 import argparse
+import asyncio
 import configparser
 import datetime
 from dateutil import parser as dateparser
@@ -15,6 +16,7 @@ import types
 import numpy as np
 import pandas as pd
 from pandas.io.common import get_filepath_or_buffer
+
 try:
     from pandas.io.common import _get_handle as get_handle
 except ImportError:
@@ -431,7 +433,7 @@ def is_function(f):
 
 def is_pandas(o):
     """Test if an object is a Pandas object"""
-    return isinstance(o, (pd.DataFrame, pd.Series, pd.Panel))
+    return isinstance(o, (pd.DataFrame, pd.Series))
 
 
 def is_file_obj(o):
@@ -512,6 +514,7 @@ def listify(o):
 
 # -------- File utils
 
+
 def open_filepath_or_buffer(f, open_flags="r", compression=None):
     """Use pandas IO functions to return a handle from a filepath
     or buffer.
@@ -535,20 +538,13 @@ def open_filepath_or_buffer(f, open_flags="r", compression=None):
         A flag indicating whether the caller should close the file object when done
 
     """
-    f, _, compression, should_close = get_filepath_or_buffer(
-        f,
-        compression=compression
-    )
+    f, _, compression, should_close = get_filepath_or_buffer(f, compression=compression)
 
     close = False or should_close
     if isinstance(f, str):
         close = True
 
-    f, handles = get_handle(
-        f,
-        open_flags,
-        compression=compression
-    )
+    f, handles = get_handle(f, open_flags, compression=compression)
 
     return f, handles, close
 
@@ -616,6 +612,42 @@ def save_excel(f, data, **kwargs):
         save_xlsx(f, data, **kwargs)
 
 
+# -------- Asyncio utils
+
+
+def get_or_create_event_loop():
+    """Get an existing asyncio Event Loop or create one if necessary"""
+    created = False
+    try:
+        loop = asyncio.get_running_loop()
+    except RuntimeError:
+        loop = asyncio.get_event_loop()
+        if loop.is_closed():
+            loop = asyncio.new_event_loop()
+            asyncio.set_event_loop(loop)
+            created = True
+    return loop, created
+
+
+def cancel_asyncio_tasks(tasks, loop, cancel_timeout=None):
+    """Cancel a set of asyncio tasks
+
+    Parameters
+    ----------
+    tasks : iterable
+        An iterable of asyncio tasks to cancel
+    loop
+        asyncio Event Loop
+    cancel_timeout : int or float, optional
+        A timeout to use when waiting for tasks to finish cancelling
+
+    """
+    [task.cancel() for task in tasks]
+    _, not_done = loop.run_until_complete(asyncio.wait(tasks, timeout=cancel_timeout))
+    if not_done:
+        warn("%d tasks did not cancel immediately" % len(not_done))
+
+
 # -------- Logging utils
 
 
